local addonName = "AccountWideRaidProfiles"
local dbName = "AccountWideRaidProfilesDB"

local addon = LibStub("AceAddon-3.0"):NewAddon(addonName,"AceHook-3.0", "AceEvent-3.0", "AceConsole-3.0")

local CVAR_FRAME_STYLE = "useCompactPartyFrames"

local RAID_CONTAINER_WIDTH = 300
local MINIMUM_RAID_CONTAINER_HEIGHT = 86

local MINIMUM_RAID_BLOCK_WIDTH  = 72
local MAXIMUM_RAID_BLOCK_WIDTH  = 144
local MINIMUM_RAID_BLOCK_HEIGHT = 36
local MAXIMUM_RAID_BLOCK_HEIGHT = 72

function addon:OnInitialize()
   self.pendingLoad = false
   self.pendingReset = false
   self.disableOptionPanelUpdate = false
   self.inCombat = false
   self.isOptionOpen = false
   self.db = LibStub("AceDB-3.0"):New(dbName)
   local addonOption = addon:SetUpOptionsPanel(addonOption)
   LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, addonOption, {"awrp"})
   self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, addonName)
   LoadAddOn("Blizzard_CompactRaidFrames")
   LoadAddOn("Blizzard_CUFProfiles")
end

function addon:OnEnable()
   if (not self.db.profile.profiles) then 
      addon:StoreAll()
   else
      addon:LoadAll()
   end
   addon:HookAll()
   addon:AlwaysShowFrame()
end

function addon:ShowRaidFrame()
   if (not self.inCombat) then
      CompactRaidFrameContainer:Show()
      CompactRaidFrameManager:Show()
   end
end

function addon:HideRaidFrame()
   if (not self.inCombat) then
         CompactRaidFrameContainer:Hide()
         CompactRaidFrameManager:Hide()
   end
end

function addon:OnOptionShow()
   self.isOptionOpen = true
   if (not self.inCombat) then
      SetCVar(CVAR_FRAME_STYLE, 1)
      addon:ShowRaidFrame()
   end
end

function addon:OnOptionHide()
   self.isOptionOpen = false
   if (not IsInGroup()) and (not self.inCombat) then
      SetCVar(CVAR_FRAME_STYLE, self.db.profile.frameStyle)
      addon:HideRaidFrame()
   end
end

function addon:AlwaysShowFrame()
   if (not self.inCombat) then
      if not IsInGroup() and self.db.profile.alwaysShowRaidFrame
            and self.db.profile.alwaysShowRaidFrame[addon:GetActiveRaidProfileId()] then
         addon:ShowRaidFrame()
      end
   end
end

function addon:HookAll()
   addon:UnhookAll()
   self.db.RegisterCallback(self, "OnProfileReset", "ResetProfiles")
   self.db.RegisterCallback(self, "OnProfileChanged", "RefreshProfiles")
   self.db.RegisterCallback(self, "OnProfileCopied", "RefreshProfiles")
   addon:Hook("SetRaidProfileSavedPosition","StorePosition", true)
   addon:Hook("CompactUnitFrameProfiles_ApplyProfile", "StoreAll", true)
   addon:Hook("SetCVar", "StoreFrameStyleChanges", true)
   addon:RegisterEvent("PLAYER_REGEN_DISABLED")
   addon:RegisterEvent("PLAYER_REGEN_ENABLED")
   addon:HookScript(self.optionsFrame, "OnShow", "OnOptionShow", true)
   addon:HookScript(self.optionsFrame, "OnHide", "OnOptionHide", true)
   addon:HookScript(CompactRaidFrameContainer, "OnHide", "AlwaysShowFrame", true)
   addon:HookScript(CompactRaidFrameManager  , "OnHide", "AlwaysShowFrame", true)
end

function addon:GetActiveRaidProfileId()
   for i=1, GetNumRaidProfiles() do
      if self.db.profile.lastProfile == self.db.profile.profiles[i] then
         return i
      end
   end
   return 1
end

function addon:PLAYER_REGEN_DISABLED()
   self.inCombat = true
end

function addon:PLAYER_REGEN_ENABLED()
   self.inCombat = false
   if self.pendingLoad then
      addon:LoadAll()
   end
   if self.pendingReset then
      addon:ResetProfiles()
   end
end

function addon:UpdateFrame()
   if (not self.inCombat) then
      CompactRaidFrameManagerDisplayFrameLockedModeToggle:SetText(UNLOCK)
      CompactRaidFrameManagerDisplayFrameLockedModeToggle.lockMode = false
      CompactRaidFrameManager_UpdateContainerLockVisibility(CompactRaidFrameManager)
      CompactRaidFrameContainer_UpdateDisplayedUnits(CompactRaidFrameContainer)
      CompactRaidFrameContainer_TryUpdate(CompactRaidFrameContainer)
      RaidOptionsFrame_UpdatePartyFrames() 
      CompactRaidFrameManager_UpdateShown(CompactRaidFrameManager)
      if (self.isOptionOpen) then
         addon:ShowRaidFrame()
      end
   end
end

function addon:RefreshProfiles()
   addon:UnhookAll()
   if (not self.db.profile.profiles) then
      addon:ResetProfiles()
   else
      addon:LoadAll()
   end
   addon:HookAll()
end

function addon:ResetProfiles()
   if not self.inCombat then
      self.pendingReset = false
      addon:UnhookAll()
      CompactUnitFrameProfiles_ResetToDefaults()
      SetCVar(CVAR_FRAME_STYLE, false)
      addon:UpdateFrame()
      addon:StoreAll()
      addon:HookAll()
   else
      addon:Print("Cannot reset profile in combat. Reset will be delayed until combat ends.")
      self.pendingReset = true
   end
end

----------------------- Load and store raid prorfiles ---------------------------

function addon:LoadAll(...)
   if not self.inCombat then
      self.pendingLoad = false
      addon:UnhookAll()
      local profiles = {}
      for i=1, GetNumRaidProfiles() do
         tinsert(profiles, GetRaidProfileName(i))
      end
      for i=1, #profiles do
         DeleteRaidProfile(profiles[i])
      end
      
      SetCVar(CVAR_FRAME_STYLE, self.db.profile.frameStyle)
      
      profiles = self.db.profile.profiles
      local options = self.db.profile.options
      local positions = self.db.profile.positions
      
      for i=1, #profiles do
         local profile = profiles[i]
         CreateNewRaidProfile(profile)
         for option, value in pairs(options[i]) do
            SetRaidProfileOption(profile, option, value)
         end
         SetRaidProfileSavedPosition(profile, unpack(positions[i]))
      end
      
      CompactUnitFrameProfiles_ActivateRaidProfile(self.db.profile.lastProfile)
      
      addon:UpdateFrame()
      addon:HookAll()
   else
      addon:Print("Cannot switch profile in combat. Switch will be delayed until combat ends.")
      self.pendingLoad = true
   end
end

function addon:StoreFrameStyleChanges(glStr, value)
   if glStr == CVAR_FRAME_STYLE and not self.isOptionOpen then
      self.db.profile.frameStyle = value
   end
end

function addon:StorePosition(...)
   args = {...}
   for i=1, #self.db.profile.positions do
      if self.db.profile.profiles[i] == args[1] then
         self.db.profile.positions[i] = {select(2,unpack(args))}
      end
   end
   if (not self.disableOptionPanelUpdate) then
      addon:UpdateOptionsPanel()
   end
end

function addon:StoreAll(...)
   local profiles = {}
   for i=1, GetNumRaidProfiles() do
      tinsert(profiles, GetRaidProfileName(i))
   end
   
   self.db.profile.frameStyle = GetCVar(CVAR_FRAME_STYLE)
   self.db.profile.profiles = profiles
   self.db.profile.options = {}
   self.db.profile.positions = {}
   
   for i=1, #profiles do
      local profile = profiles[i]
      self.db.profile.options[i] = GetRaidProfileFlattenedOptions(profile)
      self.db.profile.positions[i] = {GetRaidProfileSavedPosition(profile)}
   end
   
   self.db.profile.lastProfile = GetActiveRaidProfile()

   if (not self.disableOptionPanelUpdate) then
      addon:UpdateOptionsPanel()
   end
end

---------------------------- Option panel ----------------------------------------------
function addon:UpdateOptionsPanel()
   local addonOption = addon:SetUpOptionsPanel(addonOption)
   LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, addonOption, {"awrp", addonName})
   LibStub("AceConfigRegistry-3.0"):NotifyChange(addonName)
end

function addon:SetUpOptionsPanel()
   local option = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
   option.name = "Account Wide Raid Profiles"
   option.desc = "Manage Profiles and extra options of Account Wide Raid Profiles"
   option.args.desc.name = "The characters with the same addon profile share the same set of raid profiles."
                          .." You can adjust the active raid profile precisely by entering number at the bottom."
                          .." The raid frame is always shown when the option panel is opened.\n"
   option.args.descreset.name = ""
   option.args.rp =  {
      type = "execute",
      name = "rp",
      desc = "Open blizzard raid profile settings.",
      guiHidden = true,
      func = function()
         InterfaceOptionsFrame_OpenToCategory(_G["CompactUnitFrameProfiles"])
         end
   }
   option.args.config =  {
      type = "execute",
      name = "config",
      desc = "Open addon config panel.",
      guiHidden = true,
      func = function()
         InterfaceOptionsFrame_OpenToCategory(addonName)
         InterfaceOptionsFrame_OpenToCategory(addonName)
         end
   }
   option.args.space0 = {
    name = "",
    type = 'description',
    width = 'full',
    cmdHidden = true,
    order = 100,
   }
   option.args.configButton = {
      order = 110,
      type = "execute",
      name = "Open Raid Profile Option",
      desc = "Open Blizzard raid profile option.",
      width = "double",
      cmdHidden = true,
      func = function()
         InterfaceOptionsFrame_OpenToCategory(_G["CompactUnitFrameProfiles"])
      end
   }
   option.args.space1 = {
      name = "",
      type = 'description',
      width = 'full',
      order = 111,
   }
   option.args.isAttached = 
   {
      order = 112,
      type = "toggle",
      name = "Attach frame",
      desc = "If Raid frame attaches to the raid panel on the left edge of the screen.",
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.alwaysShowRaidFrame =
   {
      order = 113,
      type = "toggle",
      name = "Show frame when solo",
      desc = "Show raid frame even if you are not in a group. This setting is raid profile specific.",
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.space2 = {
    name = "",
    type = 'description',
    width = 'full',
    cmdHidden = true,
    order = 114,
   }
   option.args.xOffset = {
      order = 115,
      name = "x Offset",
      desc = "Pixels between the left of the screen and the topleft corner of the frame container.",
      type = "range",
      min = 0,
      step = 1,
      disabled = function(...) return addon:GetOption("isAttached") end,
      max = math.floor(GetScreenWidth()-RAID_CONTAINER_WIDTH),
      width = "normal",
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.yOffset = {
      order = 120,
      name = "y Offset",
      desc = "Pixels between the bottom of the screen and the topleft corner of the frame container.",
      type = "range",
      min = MINIMUM_RAID_CONTAINER_HEIGHT,
      step = 1,
      disabled = function(...) return addon:GetOption("isAttached") end,
      max = math.floor(GetScreenHeight()),
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.spacer3 = {
    name = "",
    type = 'description',
    width = 'full',
    cmdHidden = true,
    order = 125,
   }
   option.args.frameHeight = {
      order = 130,
      name = "Frame Height",
      desc = "Height of the frame block.",
      type = "range",
      min = MINIMUM_RAID_BLOCK_HEIGHT,
      step = 1,
      max = MAXIMUM_RAID_BLOCK_HEIGHT,
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.frameWidth = {
      order = 135,
      name = "Frame Width",
      desc = "Width of the frame block.",
      type = "range",
      min = MINIMUM_RAID_BLOCK_WIDTH,
      step = 1,
      max = MAXIMUM_RAID_BLOCK_WIDTH,
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   option.args.space4 = {
    name = "",
    type = 'description',
    width = 'full',
    cmdHidden = true,
    order = 137,
   }
   option.args.containerHeight = {
      order = 140,
      name = "Container Height",
      desc = "Pixels of the height of the frame container.",
      type = "range",
      min = MINIMUM_RAID_CONTAINER_HEIGHT,
      step = 1,
      max = math.floor(math.min(GetScreenHeight()-90,addon:GetOption("yOffset"))),
      get = function(info) return addon:GetOption(info[#info]) end,
      set = function(info, val) addon:SetOption(info[#info], val) end
   }
   return option
end

function addon:GetOption(optionName)

   local manager = CompactRaidFrameManager
   local top = manager.containerResizeFrame:GetTop()
   local bottom = manager.containerResizeFrame:GetBottom()
   local isAttached = (select(2, manager.containerResizeFrame:GetPoint(1)) == manager)
   local left = manager.containerResizeFrame:GetLeft()
   
   if (optionName == "alwaysShowRaidFrame") then
      if (not self.db.profile.alwaysShowRaidFrame) or
         (not self.db.profile.alwaysShowRaidFrame[addon:GetActiveRaidProfileId()]) then
         return false
      else
         return true
      end
   elseif (optionName == "isAttached") then
      return isAttached
   elseif (optionName == "xOffset") then
      return math.floor(left)
   elseif (optionName == "yOffset") then
      return math.floor(top)
   elseif (optionName == "containerHeight") then
      return math.floor(top-bottom)
   elseif (optionName == "frameWidth") then
      return self.db.profile.options[addon:GetActiveRaidProfileId()]["frameWidth"]
   elseif (optionName == "frameHeight") then
      return self.db.profile.options[addon:GetActiveRaidProfileId()]["frameHeight"]
   end
end

function addon:SetOption(optionName, value)
   if not self.inCombat then
      self.disableOptionPanelUpdate = true

      if (optionName == "alwaysShowRaidFrame") then
         if (not self.db.profile.alwaysShowRaidFrame) then
            self.db.profile.alwaysShowRaidFrame  = {}
         end
         self.db.profile.alwaysShowRaidFrame[addon:GetActiveRaidProfileId()] = value
         if value then
            SetCVar(CVAR_FRAME_STYLE, 1)
         end
      elseif (optionName == "frameWidth") then
         SetRaidProfileOption(GetActiveRaidProfile(), "frameWidth", value)
         CompactUnitFrameProfiles_ApplyCurrentSettings()
      elseif (optionName == "frameHeight") then
         SetRaidProfileOption(GetActiveRaidProfile(), "frameHeight", value)
         CompactUnitFrameProfiles_ApplyCurrentSettings()
      else
         local manager = CompactRaidFrameManager
         if (optionName == "isAttached") then
            manager.dynamicContainerPosition = value
         end
         if ( manager.dynamicContainerPosition ) then
            SetRaidProfileSavedPosition(GetActiveRaidProfile(), true)
         else
            local topPoint, topOffset
            local bottomPoint, bottomOffset
            local leftPoint, leftOffset
            
            local screenHeight = GetScreenHeight()
            local top = manager.containerResizeFrame:GetTop()
            local topDifference = 0
            if (optionName == "yOffset") then
               topDifference  = value - top        
               top = value
            end
            if ( top > screenHeight / 2 ) then
               topPoint = "TOP"
               topOffset = screenHeight - top
            else
               topPoint = "BOTTOM"
               topOffset = top
            end
            
            local bottom = manager.containerResizeFrame:GetBottom()
            if (optionName == "yOffset") then
               bottom = bottom + topDifference
               if (bottom < 0) then
                  bottom = 0
               end
            elseif (optionName == "containerHeight") then
               bottom = top - value
            end
            if ( bottom > screenHeight / 2 ) then
               bottomPoint = "TOP"
               bottomOffset = screenHeight - bottom
            else
               bottomPoint = "BOTTOM"
               bottomOffset = bottom
            end
            
            local isAttached = (select(2, manager.containerResizeFrame:GetPoint(1)) == manager)
            if (optionName == "isAttached") then
              isAttached = value
            end
            if ( isAttached ) then
               leftPoint = "ATTACHED"
               leftOffset = 0
            else
               local screenWidth = GetScreenWidth()
               local left = manager.containerResizeFrame:GetLeft()
               if (optionName == "xOffset") then
                  left = value
               end
               if ( left > screenWidth / 2 ) then
                  leftPoint = "RIGHT"
                  leftOffset = screenWidth - left
               else
                  leftPoint = "LEFT"
                  leftOffset = left
               end
            end
      
            SetRaidProfileSavedPosition(GetActiveRaidProfile(), false
               , topPoint, topOffset, bottomPoint, bottomOffset, leftPoint, leftOffset)
         end
         CompactRaidFrameManager_ResizeFrame_LoadPosition(CompactRaidFrameManager)
      end
      self.disableOptionPanelUpdate = false
      addon:UpdateFrame()
      addon:UpdateOptionsPanel()
   else
      addon:Print("Cannot change options in combat.")
   end
end